"""
Git IZI : AI
"""

from json import JSONDecodeError

from aiohttp import ClientOSError, WSServerHandshakeError

from gitizi.util import suppress_output

from ..exceptions import GitiziException, GitiziParseException

with suppress_output():
    import g4f

from g4f.Provider import Bing  # pylint:disable=wrong-import-order

from .context import Context
from .context.squash import CONTEXT as S_CONTEXT
from .context.squash import USER_MSG_WRAPPER as S_WRAPPER
from .context.squash import retrieve_response as S_RESPONSE


def contextualize(msg: str, ctx: Context = Context.ASK) -> list[dict]:
    """Contextualize message"""
    match ctx:
        case Context.SQUASH:
            ai_ctx = S_CONTEXT
            wrapper = S_WRAPPER
            msg = msg if msg else "No changes"
        case Context.ASK | _:
            ai_ctx = []
            wrapper = "{user_message}"
    ai_ctx.append({"role": "user", "content": wrapper.format(user_message=msg)})

    return ai_ctx


def parse_response(msg, ctx: Context = Context.ASK) -> str:
    """Return contextualized response message"""
    match ctx:
        case Context.SQUASH:
            return S_RESPONSE(response=msg)
        case Context.ASK | _:
            return msg


def ask(msg: str, ctx: Context = Context.ASK):
    """Ask AI"""
    messages = contextualize(msg=msg, ctx=ctx)

    with suppress_output():
        for _ in range(3):
            try:
                response = g4f.ChatCompletion.create(
                    model=g4f.models.default,
                    messages=messages,
                    provider=Bing,
                    auth=True,
                )
                return parse_response(msg=response, ctx=ctx)
            except (WSServerHandshakeError, ClientOSError, GitiziParseException, ConnectionResetError):
                pass
        raise GitiziException("Connection error")



"""
Git IZI : Util
"""

import os
import sys
from contextlib import contextmanager, redirect_stderr, redirect_stdout

import git
import click

from .exceptions import GitiziException


@contextmanager
def suppress_output():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(os.devnull, "w", encoding="utf-8") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield err, out


try:
    repo = git.Repo(os.getcwd())
except git.exc.InvalidGitRepositoryError:
    click.echo("Invalid repository!")
    sys.exit(1)


def default_remote():
    """Get default remote"""
    candidates = ["origin", "remote"]
    for remote in repo.remotes:
        if remote.name in candidates:
            return remote
    return repo.remotes[0]


def default_remote_branch(remote: str | git.Remote | None = None) -> git.Reference:
    """Get default branch"""
    remote = (
        remote if isinstance(remote, git.Remote) else git.Remote(repo=repo, name=remote) if remote else default_remote()
    )

    return next(filter(lambda ref: ref.name == f"{remote.name}/HEAD", remote.refs)).ref


def default_local_branch() -> git.Reference:
    """Get default local branch"""
    remote_branch = default_remote_branch()

    return next(
        filter(
            lambda branch: branch.tracking_branch() == remote_branch,
            remote_branch.repo.branches,
        )
    )


def current_branch() -> git.Head:
    """Get current branch"""
    return repo.active_branch


def is_current_default() -> bool:
    """Check if current branch is a default branch"""
    current = current_branch()
    default = default_local_branch()

    return current == default


def default_synced(can_behind: bool = False, can_ahead: bool = False) -> bool:
    """Check if default branch is synced with remote"""
    local = default_local_branch()
    remote = default_remote_branch()

    if not can_behind:
        commits_behind = repo.iter_commits(f"{local.name}..{remote.name}")
        count_behind = sum(1 for _ in commits_behind)
    else:
        count_behind = 0

    if not can_ahead:
        commits_ahead = repo.iter_commits(f"{remote.name}..{local.name}")
        count_ahead = sum(1 for _ in commits_ahead)
    else:
        count_ahead = 0

    return count_behind == 0 and count_ahead == 0


def current_synced(can_behind: bool = False, can_ahead: bool = False) -> bool:
    """Check if current branch is in sync with default local branch"""
    curr_branch = current_branch()

    local_name = default_local_branch().name

    if not can_behind:
        commits_behind = repo.iter_commits(f"{curr_branch.name}..{local_name}")
        count_behind = sum(1 for _ in commits_behind)
    else:
        count_behind = 0

    if not can_ahead:
        commits_ahead = repo.iter_commits(f"{local_name}..{curr_branch.name}")
        count_ahead = sum(1 for _ in commits_ahead)
    else:
        count_ahead = 0

    return count_behind == 0 and count_ahead == 0


def reset_default():
    """Fully resets default branch to be in-line with its remote"""
    repo.head.reset(commit=f"{default_remote()}/HEAD", index=True, working_tree=True, paths=None)


def all_commited() -> bool:
    """Check if everything is commited and we can safely check out"""
    return not repo.is_dirty()


def checkout(branch: str | None = None, new: bool = False) -> tuple[bool, str]:
    """Checkout to provided branch or to default branch if None"""
    if branch is None:
        branch = default_local_branch().name
    else:
        branches = (branch.name for branch in repo.branches)
        if new:
            if branch in branches:
                raise GitiziException("Branch already exists")
        else:
            if branch not in branches:
                raise GitiziException("Incorrect branch")

    curr_branch = current_branch().name
    if branch != curr_branch:
        args = tuple()
        if new:
            args = args + ("-b",)
        args = args + (branch,)
        repo.git.checkout(*args)
        return True, curr_branch
    return False, curr_branch


def fetch_default():
    """Fetch default branch from default remote"""
    default_remote().fetch(refspec=f"refs/heads/{default_local_branch().name}:{default_remote_branch().path}")


def commit(with_no_edit: bool = False):
    """Create a commit with all uncommited changes"""
    files = repo.git.diff(None, name_only=True)
    for f in files.split("\n"):
        if f:
            repo.git.add(f)
    if not with_no_edit:
        repo.git.commit(
            "-m",
            "Modified files: " + ", ".join(files.split("\n")),
            "--allow-empty",
        )
    else:
        try:
            repo.git.commit("--no-edit", "--allow-empty")
        except git.exc.GitCommandError:
            # Empty commit message (no merge was required)
            pass


def merge_to_current(
    branch: str = None,
    with_squash: bool = False,
    with_commit: bool = False,
    with_no_edit: bool = False,
):
    """Merge default branch to current branch"""
    branch = default_local_branch().name if branch is None else branch

    args = (branch,)
    if with_squash:
        args = args + ("--squash",)
    if with_commit:
        args = args + ("--commit",)
    if with_no_edit:
        args = args + ("--no-edit",)

    repo.git.merge(*args)


def get_hash():
    """Get current commit hash"""
    return repo.head._get_commit().hexsha  # pylint:disable=protected-access


def last_shared():
    """Get last shared commit hash"""
    if is_current_default():
        return None
    branch = default_remote_branch().name
    cmd = ["git", "merge-base", "--fork-point", branch]
    last_shared_hash = repo.git.execute(cmd)
    return last_shared_hash


def get_log_until(commit_hash: str):
    """Get commit messages until commit_hash"""
    cmd = ["git", "log", f"{commit_hash}..", "--pretty=format:%s", "--abbrev-commit"]
    return repo.git.execute(cmd)


def reset_branch(from_hash: str | None = None):
    """Reset current branch"""
    if from_hash:
        repo.head.reset(commit=from_hash)
    else:
        repo.head.reset()


def delete_branch(branch_name: str):
    """Delete a branch"""
    cmd = ["git", "branch", "-D", branch_name]
    repo.git.execute(cmd)


def ammend_commit_msg(new_msg: str):
    """Ammend last commit message"""
    repo.git.commit("--amend", "-m", new_msg, "--allow-empty")



"""
Git IZI : CLI
"""

import sys

import click

from .ai import Context
from .ai import ask as ask_ai
from .util import (
    commit,
    checkout,
    get_hash,
    last_shared,
    all_commited,
    reset_branch,
    delete_branch,
    fetch_default,
    get_log_until,
    reset_default,
    current_branch,
    current_synced,
    default_remote,
    default_synced,
    merge_to_current,
    ammend_commit_msg,
    is_current_default,
    default_local_branch,
    default_remote_branch,
)
from .exceptions import GitiziException


@click.group()
def main():
    """Git IZI entrypoint command"""


def _default(reset: bool):
    """Helper function"""
    fetch_default()
    click.echo(
        f"[{default_local_branch().name}] ----[In sync: {default_synced()}]----> [{default_remote_branch().name}]"
    )

    if reset:
        checked = False
        old_branch = None
        if not is_current_default():
            if not all_commited():
                click.echo("All local changes have to be commited or stashed before reset")
                raise click.Abort
            checked, old_branch = checkout()
        if not default_synced():
            reset_default()
            if checked:
                checkout(old_branch)
            click.echo(f"Default branch [{default_local_branch()}] synchronised")
        else:
            if checked:
                checkout(old_branch)
            click.echo(f"Branch [{default_local_branch()}] is already synchronised")


@main.command()
@click.option(
    "--reset/--no-reset",
    is_flag=True,
    default=False,
    help="Reset local branch to its remote state.",
)
def default(reset: bool = False):
    """Print default branch name"""
    _default(reset=reset)


@main.group()
def current():
    """Operations on current branch"""


@current.command()
def info():
    """Current branch info"""
    click.echo(f"Current branch is: {current_branch()}, is default: {is_current_default()}")
    click.echo(
        f"Default branch is: {default_local_branch()}, is in sync with remote [{default_remote()}]: {default_synced()}"
    )


def _squash(force: bool = False):
    """Squash current commit"""
    fetch_default()
    if is_current_default():
        click.echo(f"Branch currenly in use is a default branch: {current_branch()}")
        raise click.Abort
    click.echo("Non-default branch, continuing...")

    if not all_commited():
        if not force:
            click.echo("All local changes have to be commited or stashed before squash")
            raise click.Abort
        commit()
    click.echo("Everything is commited, continuing...")

    if not default_synced():
        if not force:
            click.echo(
                f"Default branch [{default_local_branch()}] is not in sync with "
                f"remote default [{default_remote_branch()}]"
            )
            raise click.Abort
        _default(reset=True)
        # TODO: Pull default -- check if reset is all that's needed  pylint:disable=fixme
    click.echo("Default branch is in sync with remote, continuing...")

    if not current_synced(can_ahead=True):
        if not force:
            click.echo(f"Current branch [{current_branch()}] is behind with local default [{default_local_branch()}]")
            raise click.Abort
        merge_to_current()
    click.echo("Default branch is ahead or in sync with default, continuing...")

    shared = last_shared()
    logs = get_log_until(commit_hash=shared)

    _, old_branch = checkout()  # Checkout to default
    checkout(branch="gitizi/temp", new=True)

    merge_to_current(branch=old_branch, with_squash=True)
    commit(with_no_edit=True)
    current_hash = get_hash()

    checkout(branch=old_branch)
    reset_branch(from_hash=current_hash)

    delete_branch("gitizi/temp")
    click.echo("Commits squashed, aggregating message...")

    try:
        squashed_message = ask_ai(msg=logs, ctx=Context.SQUASH)
        ammend_commit_msg(squashed_message)
        click.echo(f"Amended commit message: {squashed_message}")
        sys.exit(0)
    except GitiziException:
        click.echo("Connection failure. Generating squashed commit message failed.")
        sys.exit(1)


@current.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Try to perform all actions needed to create a squashed commit.",
)
def squash(force: bool = False):
    """Command to squash, optionally forceful"""
    _squash(force=force)


@main.command()
@click.argument("message")
def ask(message: str):
    """Ask AI"""
    click.echo(ask_ai(message))


@click.command()
def main_squash():
    """Command to squash with force - please know what you're doing"""
    _squash(force=True)



"""
Git IZI : Squash
"""

import re
import json
from json import JSONDecodeError

from gitizi.exceptions import GitiziParseException

CONTEXT = [
    {
        "role": "system",
        "content": (
            "You are a GitHub plugin that's responsible for parsing squashed commit messages and rewriting them"
            " into a simple one-liners. You also detect whether or not there was a version bump within the squashed"
            " messages.If you do detect a version bump, you also prepend an additional square bracket tag that "
            " looks like this: [Bump <previous version> → <new version>]. If you can't detect <previous version>"
            " then use [Bump <new version>] format. Remember that squashed commits can be nested. That is: oldest"
            " commits are at the very bottom and have the biggest indent. Remove redundant information like"
            " lint and unit test fixes, unless that's the only information provided. If you do not detect the"
            " squash - merge multilined messages into a singular line. Message content's are contained within <message>"
            ' tags. Return a JSON message containing a summarised message in a following format: {"message":'
            " <summarized commit message>}. Nothing more. Otherwise there'll be trouble. Provided user commit messages"
            " are ALWAYS valid. Even if you think otherwise."
        ),
    },
    {
        "role": "user",
        "content": "Squashed commit message:\n <message>Initialize repository.\nFix linting issues.\nAdd unit tests."
        "</message>",
    },
    {
        "role": "assistant",
        "content": '{"message":"Initizalize repo and add unit tests."}',
    },
]

USER_MSG_WRAPPER = "Squashed commit message:\n <message>{user_message}</message>"


def retrieve_response(response: str):
    """Retrieve response from AI"""
    try:
        message: str = json.loads(response)["message"]
        if not message:
            raise GitiziParseException
        if not message[0].isalpha() and not re.match(r"^\[Bump .*\].*$", message):
            raise GitiziParseException
        return message
    except JSONDecodeError as exc:
        raise GitiziParseException from exc
